import json

import six
from kubernetes.client import ApiClient


def list_namespaced_resources(kind, api_group, namespace, **kwargs):
  local_var_params = locals()

  all_params = [
    'namespace',
    'field_selector',
    'label_selector',
    'resource_version'
  ]

  for key, val in six.iteritems(local_var_params['kwargs']):
    local_var_params[key] = val

  path_params = {'namespace': local_var_params['namespace']}

  del local_var_params['kwargs']

  query_params = []
  if 'field_selector' in local_var_params:
    query_params.append(('fieldSelector', local_var_params['field_selector']))  # noqa: E501
  if 'label_selector' in local_var_params:
    query_params.append(('labelSelector', local_var_params['label_selector']))  # noqa: E501
  if 'resource_version' in local_var_params:
    query_params.append(('resourceVersion', local_var_params['resource_version']))  # noqa: E501

  header_params = {}

  form_params = []
  local_var_files = {}

  api_client = ApiClient()

  body_params = None
  header_params['Accept'] = api_client.select_header_accept([
    'application/json',
    'application/yaml',
    'application/vnd.kubernetes.protobuf',
    'application/json;stream=watch',
    'application/vnd.kubernetes.protobuf;stream=watch'
  ])

  auth_settings = ['BearerToken']  # noqa: E501
  url = f"{request_sig(api_group)}/namespaces/{{namespace}}/{kind}"

  response = api_client.call_api(
    url,
    'GET',
    path_params,
    query_params,
    header_params,
    body=body_params,
    post_params=form_params,
    files=local_var_files,
    auth_settings=auth_settings,
    async_req=local_var_params.get('async_req'),
    _return_http_data_only=True,
    _preload_content=False,
    _request_timeout=local_var_params.get('_request_timeout'),
    collection_formats={}
  )

  reg = json.loads(response.data.decode('utf-8'))
  inferred_kind = reg['kind'].replace('List', '')
  for item in reg['items']:
    item['kind'] = inferred_kind
  return reg


def delete_namespaced_resource(kind, api_group, namespace, name, **kwargs):
  local_var_params = locals()

  all_params = [
    'namespace',
    'resource_version'
  ]

  for key, val in six.iteritems(local_var_params['kwargs']):
    local_var_params[key] = val

  path_params = {
    'name': local_var_params['name'],
    'namespace': local_var_params['namespace']
  }

  del local_var_params['kwargs']

  query_params = []
  if 'resource_version' in local_var_params:
    query_params.append(('resourceVersion', local_var_params['resource_version']))

  header_params = {}

  api_client = ApiClient()

  body_params = None
  header_params['Accept'] = api_client.select_header_accept([
    'application/json',
    'application/yaml',
    'application/vnd.kubernetes.protobuf',
    'application/json;stream=watch',
    'application/vnd.kubernetes.protobuf;stream=watch'
  ])

  url = f"{request_sig(api_group)}/namespaces/{{namespace}}/{kind}/{{name}}"

  response = api_client.call_api(
    url,
    'DELETE',
    path_params,
    query_params,
    {},
    body=body_params,
    post_params=[],
    files={},
    auth_settings=['BearerToken'],
    async_req=local_var_params.get('async_req'),
    _return_http_data_only=True,
    _preload_content=False,
    _request_timeout=local_var_params.get('_request_timeout'),
    collection_formats={}
  )

  return response.status == 200


def request_sig(api_group: str):
  api_group = '' if api_group == 'v1' else api_group
  prefix = 'api' if api_group == '' else 'apis'
  version = api_group.split('/')[1] if '/' in api_group else 'v1'
  api_group = api_group.split('/')[0] if '/' in api_group else api_group
  parts = [s for s in [prefix, api_group, version] if s]
  return f"/{'/'.join(parts)}"



from kubernetes.client import V1ObjectMeta, V1PodSpec, V1Container, V1ResourceRequirements

from k8kat.auth.kube_broker import broker

def pod(**subs):
  default_labels = dict(app=subs['name'])
  labels = {**subs.get('labels', {}), **default_labels}

  resource_reqs = None
  container_resources = subs.get('resources', None)
  if container_resources:
    resource_reqs = V1ResourceRequirements(**container_resources)

  return broker.client.V1Pod(
    metadata=V1ObjectMeta(
      name=subs.get('name'),
      labels=labels
    ),
    spec=V1PodSpec(
      restart_policy=subs.get('restart', 'Always'),
      containers=[
        V1Container(
          name=subs.get('container', 'primary'),
          image=subs.get('image', 'nginx'),
          image_pull_policy="IfNotPresent",
          command=subs.get('command'),
          args=subs.get('args'),
          resources=resource_reqs
        )
      ]
    )
  )


def create(**subs):
  return broker.coreV1.create_namespaced_pod(
    body=pod(**subs),
    namespace=subs['ns']
  )



from functools import lru_cache
from typing import List

from kubernetes.client import V1Namespace

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes, MetricsDict
from k8kat.res.pod.kat_pod import KP
from k8kat.res.relation.relation import Relation
from k8kat.res.sa.kat_service_account import KatServiceAccount
from k8kat.utils.main.class_property import classproperty


class KatNs(KatRes):

  @classproperty
  def kind(self):
    return "Namespace"

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  def _perform_delete_self(self):
    broker.coreV1.delete_namespace(self.name)

  def is_active(self) -> bool:
    if self.raw.status:
      return self.raw.status.phase == 'Active'
    else:
      return False

  def is_work_ready(self) -> bool:
    if self.is_active():
      default_sa = KatServiceAccount.find('default', self.name)
      if default_sa:
        if len(default_sa.secrets()) == 1:
          return None not in default_sa.secrets()
    return False

  @lru_cache(maxsize=128)
  def load_per_pod_metrics(self) -> List[MetricsDict]:
    """Loads the appropriate metrics dict from k8s metrics API."""
    return broker.custom.list_namespaced_custom_object(
      group='metrics.k8s.io',
      version='v1beta1',
      namespace=self.name,
      plural='pods'
    )['items']

# --
# --
# --
# -------------------------------PLUMBING-------------------------------
# --
# --
# --

  def body(self) -> V1Namespace:
    return self.raw

  @classmethod
  def is_namespaced(cls) -> bool:
    return False

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespace,
      list=broker.coreV1.list_namespace,
      patch=broker.coreV1.patch_namespace,
      delete=broker.coreV1.delete_namespace
    )

# --
# --
# --
# -------------------------------RELATIONS-------------------------------
# --
# --
# --

  def pods(self, **query) -> List[KP]:
    """Selects and returns pods associated with the namespace."""
    from k8kat.res.pod.kat_pod import KatPod
    return Relation[KatPod](
      model_class=KatPod,
      ns=self.name,
      **query
    )



from kubernetes.client import V1Secret

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.class_property import classproperty


class KatSecret(KatRes):

  def raw_ob(self) -> V1Secret:
    return self.raw

  @classproperty
  def kind(self):
    return "Secret"

  @classmethod
  def list_excluding_sys(cls, ns=None, **query):
    updated_query = dict(
      **query,
      not_fields={
        **(query.get('not_fields', {})),
        'type': 'kubernetes.io/service-account-token'
      }
    )
    return cls.list(ns, **updated_query)

  @classmethod
  def k8s_verb_methods(cls):
    return(
      dict(
        read=broker.coreV1.read_namespaced_secret,
        patch=broker.coreV1.patch_namespaced_secret,
        delete=broker.coreV1.delete_namespaced_secret,
        list=broker.coreV1.list_namespaced_secret,
      )
    )



from kubernetes.client import V1ObjectMeta, V1ServiceSpec, V1ServicePort
from k8kat.auth.kube_broker import broker


def create(**subs):
  default_labels = dict(app=subs['name'])
  labels = {**subs.get('labels', {}), **default_labels}
  match_labels = {**labels, **subs.get('selector', {})}

  svc = broker.client.V1Service(
    api_version='v1',
    metadata=V1ObjectMeta(
      name=subs.get('name'),
      labels=labels
    ),
    spec=V1ServiceSpec(
      type=subs.get('type', 'ClusterIP'),
      selector=match_labels,
      ports=[
        V1ServicePort(
          port=subs.get('from_port', 80),
          target_port=subs.get('to_port', 80)
        )
      ]
    )
  )

  return broker.coreV1.create_namespaced_service(
    body=svc,
    namespace=subs['ns']
  )



from functools import lru_cache
from typing import Dict, List

from kubernetes.client import V1Job

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes, MetricsDict
from k8kat.res.base.label_set_expressions import label_conditions_to_expr
from k8kat.res.pod.kat_pod import KP
from k8kat.res.relation.relation import Relation
from k8kat.utils.main.class_property import classproperty


class KatJob(KatRes):

  @classproperty
  def kind(self):
    return "Job"

  @property
  def pod_select_labels(self) -> Dict[str, str]:
    return self.raw.spec.selector.match_labels or {}

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  @lru_cache(maxsize=128)
  def load_per_pod_metrics(self) -> List[MetricsDict]:
    """Loads the appropriate metrics dict from k8s metrics API."""
    return broker.custom.list_namespaced_custom_object(
      group='metrics.k8s.io',
      version='v1beta1',
      namespace=self.namespace,
      label_selector=label_conditions_to_expr(self.pod_select_labels.items(), []),
      plural='pods'
    )['items']

# # --
# # --
# # --
# # -------------------------------ACTION-------------------------------
# # --
# # --
# # --

  def body(self) -> V1Job:
    return self.raw

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.batchV1.read_namespaced_job,
      patch=broker.batchV1.patch_namespaced_job,
      delete=broker.batchV1.delete_namespaced_job,
      list=broker.batchV1.list_namespaced_job
    )

# --
# --
# --
# -------------------------------RELATIONS-------------------------------
# --
# --
# --

  def pods(self, **query) -> List[KP]:
    """Selects and returns pods associated with the job."""
    from k8kat.res.pod.kat_pod import KatPod
    return Relation[KatPod](
      model_class=KatPod,
      ns=self.ns,
      labels=self.pod_select_labels,
      **query
    )




from typing import List

from kubernetes.client import V1Role, V1PolicyRule

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.class_property import classproperty

main_verbs = ['list', 'read', 'patch', 'create', 'delete']


def verbs_to_bools(rule_verbs: List[str]):
  if rule_verbs == ['*']:
    return {verb: True for verb in main_verbs}
  else:
    return {verb: (verb in rule_verbs) for verb in main_verbs}


class KatRole(KatRes):

  @classproperty
  def kind(self):
    return "Role"

  def body(self) -> V1Role:
    return self.raw

  def matrix_form(self):
    out = []
    for rule in self.body().rules:
      for resource_type in rule.resources:
        out.append(dict(
          resource=resource_type,
          apis=rule.api_groups,
          verbs=verbs_to_bools(rule.verbs)
        ))
    return out

  @classmethod
  def k8s_verb_methods(cls):
    return(
      dict(
        read=broker.rbacV1.read_namespaced_role,
        patch=broker.rbacV1.patch_namespaced_role,
        delete=broker.rbacV1.delete_namespaced_role,
        list=broker.rbacV1.list_namespaced_role,
      )
    )


class KatRoleBinding(KatRes):

  @classproperty
  def kind(self):
    return "RoleBinding"

  @classmethod
  def k8s_verb_methods(cls):
    return(
      dict(
        read=broker.rbacV1.read_namespaced_role_binding,
        patch=broker.rbacV1.patch_namespaced_role_binding,
        delete=broker.rbacV1.delete_namespaced_role_binding,
        list=broker.rbacV1.list_namespaced_role_binding
      )
    )


class KatClusterRole(KatRes):

  @classproperty
  def kind(self):
    return "ClusterRole"

  @classmethod
  def is_namespaced(cls):
    return False

  @classmethod
  def k8s_verb_methods(cls):
    return(
      dict(
        read=broker.rbacV1.read_cluster_role,
        patch=broker.rbacV1.patch_cluster_role,
        delete=broker.rbacV1.delete_cluster_role,
      )
    )


class KatClusterRoleBinding(KatRes):

  @classproperty
  def kind(self):
    return "ClusterRoleBinding"

  @classmethod
  def is_namespaced(cls):
    return False

  @classmethod
  def k8s_verb_methods(cls):
    return(
      dict(
        read=broker.rbacV1.read_cluster_role_binding,
        patch=broker.rbacV1.patch_cluster_role_binding,
        delete=broker.rbacV1.delete_cluster_role_binding,
      )
    )



from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.class_property import classproperty


class KatPvc(KatRes):

  @classproperty
  def kind(self):
    return "PersistentVolumeClaim"

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespaced_persistent_volume_claim,
      patch=broker.coreV1.patch_namespaced_persistent_volume_claim,
      delete=broker.coreV1.delete_namespaced_persistent_volume_claim,
      list=broker.coreV1.list_namespaced_persistent_volume_claim
    )



from typing import List, Dict

from kubernetes.client import ExtensionsV1beta1Ingress, ExtensionsV1beta1IngressRule, ExtensionsV1beta1HTTPIngressPath

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.class_property import classproperty


class KatIngress(KatRes):

  @classproperty
  def kind(self):
    return 'Ingress'

  def body(self) -> ExtensionsV1beta1Ingress:
    return self.raw

  def basic_rules(self) -> Dict[str, List[Dict[str, str]]]:
    result = {}
    rules: List[ExtensionsV1beta1IngressRule] = self.body().spec.rules
    for rule in rules:
      if rule.host:
        bundles = []
        ingress_paths: List[ExtensionsV1beta1HTTPIngressPath] = rule.http.paths
        for ingress_path in ingress_paths:
          bundles.append(dict(
            service=ingress_path.backend.service_name,
            port=ingress_path.backend.service_port,
            path=ingress_path.path
          ))
        result[rule.host] = bundles
    return result


  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.extsV1.read_namespaced_ingress,
      patch=broker.extsV1.patch_namespaced_ingress,
      delete=broker.extsV1.delete_namespaced_ingress,
      list=broker.extsV1.list_namespaced_ingress
    )



from typing import List

from kubernetes.client import V1ServiceAccount

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main.class_property import classproperty


class KatServiceAccount(KatRes):

  @classproperty
  def kind(self):
    return "ServiceAccount"

  def body(self) -> V1ServiceAccount:
    return self.raw

  @classmethod
  def list_excluding_sys(cls, ns=None, **query):
    updated_query = dict(
      **query,
      not_fields={
        **(query.get('not_fields', {})),
        'metadata.name': 'default',
      }
    )
    return cls.list(ns, **updated_query)

  @classmethod
  def k8s_verb_methods(cls):
    return (
      dict(
        read=broker.coreV1.read_namespaced_service_account,
        patch=broker.coreV1.patch_namespaced_service_account,
        delete=broker.coreV1.delete_namespaced_service_account,
        list=broker.coreV1.list_namespaced_service_account,
      )
    )

  def secrets(self) -> List[any]:
    from k8kat.res.secret.kat_secret import KatSecret
    make = lambda sd: KatSecret.find(sd.name, sd.namespace or self.ns)
    secret_descriptors = self.body().secrets or []
    return [make(secret_desc) for secret_desc in secret_descriptors]



from typing import List, Optional
from typing_extensions import TypedDict


class UsageDict(TypedDict):
  cpu: Optional[str]
  memory: Optional[str]


class ContainerMetricsDict(TypedDict):
  usage: UsageDict


class PodMetricsDict(TypedDict):
  containers: List[ContainerMetricsDict]


class NodeMetricsDict(TypedDict):
  usage: UsageDict




import re
import time
from threading import Thread
from typing import Tuple, List

from kubernetes.client import V1Namespace, V1ObjectMeta
from kubernetes.client.rest import ApiException

from k8kat.auth.kube_broker import broker
from k8kat.res.ns.kat_ns import KatNs

config = dict(
  max_ns=15
)

def possible_names():
  return [make_name(i) for i in range(max_ns())]


def max_ns() -> int:
  return config['max_ns']

def update_max_ns(new_max_ns):
  assert new_max_ns >= max_ns()
  config['max_ns'] = new_max_ns


def make_name(index: int) -> str:
  return f"ns{index + 1}"


def is_nectar_test_ns(name: str) -> bool:
  return re.search("^ns(\d*)$", name) is not None


def create_ns(name) -> str:
  broker.coreV1.create_namespace(
    body=V1Namespace(metadata=V1ObjectMeta(name=name))
  )

  kat_ns = KatNs.find(name)
  while not (kat_ns and kat_ns.is_work_ready()):
    kat_ns = KatNs.find(name, name)
    time.sleep(.5)
  return name


def get_ns() -> List[Tuple[str, str]]:
  api = broker.coreV1
  simplified = lambda n: (n.metadata.name, n.status.phase)
  all_cluster_ns = [simplified(ns) for ns in api.list_namespace().items]
  return [ns for ns in all_cluster_ns if is_nectar_test_ns(ns[0])]


def avail_now_names(crt_list) -> List[str]:
  crt_names = [ns[0] for ns in crt_list]
  terminating = terminating_names(crt_list)
  return list(set(possible_names()) - set(crt_names) - set(terminating))


def terminating_names(crt_list) -> List[str]:
  return [ns[0] for ns in crt_list if ns[1] == 'Terminating']


def running_names(crt_list) -> List[str]:
  return [ns[0] for ns in crt_list if ns[1] == 'Active']


def wait_for_term(wait_for_n: int):
  print(f"Waiting for {wait_for_n} namespaces to finish being destroyed...")
  while len(avail_now_names(get_ns())) < wait_for_n:
    time.sleep(1)


def destroy_ns(name):
  try:
    broker.coreV1.delete_namespace(name)
  except ApiException:
    pass


def initiate_ns_destroy(name):
  Thread(target=destroy_ns, args=(name,)).start()


def destroy_namespaces_async(terminating, count, spared: List[str]):
  victim_names = set(possible_names()) - set(terminating) - set(spared)

  if victim_names >= count:
    victim_names = victim_names[:count]
    for victim_name in victim_names:
      initiate_ns_destroy(victim_name)
  else:
    raise RuntimeError("Cluster is full!")


def request(count: int, spared: List[str] = None) -> List[str]:
  spared = spared if spared is not None else []
  crt_state = get_ns()
  avail_now = avail_now_names(crt_state)
  terminating = terminating_names(crt_state)
  if len(avail_now) >= count:
    required_subset = avail_now[:count]
    return [create_ns(name) for name in required_subset]
  else:
    missing_count = len(avail_now) - count
    if not len(terminating) >= missing_count:
      amount_needed = missing_count - len(terminating)
      destroy_namespaces_async(amount_needed, terminating, spared)
    wait_for_term(len(terminating) - missing_count)
    avail_now = avail_now_names(get_ns())
    assert len(avail_now) >= count
    return [create_ns(name) for name in avail_now]


def relinquish(*names: List[str]):
  for name in names:
    if name not in terminating_names(get_ns()):
      initiate_ns_destroy(name)


def relinquish_all():
  relinquish(*running_names(get_ns()))




def kat_classes():

  from k8kat.res.config_map.kat_map import KatMap
  from k8kat.res.dep.kat_dep import KatDep
  from k8kat.res.ns.kat_ns import KatNs
  from k8kat.res.pod.kat_pod import KatPod
  from k8kat.res.pvc.kat_pvc import KatPvc
  from k8kat.res.rbac.rbac import KatRole, KatRoleBinding, KatClusterRole, KatClusterRoleBinding
  from k8kat.res.sa.kat_service_account import KatServiceAccount
  from k8kat.res.secret.kat_secret import KatSecret
  from k8kat.res.svc.kat_svc import KatSvc
  from k8kat.res.ingress.kat_ingress import KatIngress
  from k8kat.res.quotas.kat_quota import KatQuota
  from k8kat.res.node.kat_node import KatNode
  from k8kat.res.job.kat_job import KatJob
  from k8kat.res.rs.kat_rs import KatRs

  return [
    KatRole,
    KatRoleBinding,
    KatClusterRole,
    KatClusterRoleBinding,
    KatServiceAccount,
    KatNs,
    KatNode,
    KatMap,
    KatSecret,
    KatPvc,
    KatPod,
    KatDep,
    KatSvc,
    KatIngress,
    KatQuota,
    KatJob,
    KatRs
  ]



import time
from datetime import datetime
from typing import Dict, Callable, Optional, Type, TypeVar, List

import inflection
import kubernetes
from cachetools.func import lru_cache
from kubernetes.client import ApiClient
from kubernetes.client.rest import ApiException

from k8kat.auth.kube_broker import broker
from k8kat.res.base import rest_backend
from k8kat.res.events.kat_event import KatEvent
from k8kat.utils.main import utils, res_utils, units
from k8kat.utils.main.api_defs_man import api_defs_man
from k8kat.utils.main.class_property import classproperty
from k8kat.utils.main.types import PodMetricsDict

MetricsDict = TypeVar('MetricsDict')
KR = TypeVar('KR', bound='KatRes')


class KatRes:

  def __init__(self, raw):
    self.is_dirty = False
    if type(raw) == dict:
      raw = from_dict(raw)
    self.raw = raw

# --
# --
# --
# -------------------------------PROPERTIES-------------------------------
# --
# --
# --

  @property
  def uid(self):
    return self.raw.metadata.uid

  @classproperty
  def kind(self) -> str:
    raise NotImplementedError

  @classproperty
  def res_name_plural(self):
    return f"{self.kind.lower()}s"

  @classproperty
  def kind_aliases(self) -> List[str]:
    return [
      self.kind,
      self.res_name_plural
    ]

  @property
  def name(self) -> str:
    return self.raw.metadata.name

  @property
  def namespace(self) -> str:
    return self.raw.metadata.namespace

  def sig(self, with_ns=False):
    ns_less = f'{self.kind}:{self.name}'
    return f'{self.ns}:{ns_less}' if with_ns else ns_less

  @property
  def ns(self) -> str:
    return self.namespace

  @property
  def metrics_component(self):
    return None

  @property
  def labels(self) -> Dict[str, str]:
    return self.raw.metadata.labels or {}

  @property
  def annotations(self) -> Dict[str, str]:
    return self.raw.metadata.annotations or {}

  @property
  def created_at(self) -> Optional[str]:
    ts = self.raw.metadata.creation_timestamp
    return ts.isoformat(' ', 'seconds') if ts else None

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  def ternary_status(self):
    return 'positive'

  def __lt__(self, other):
    return self.created_at < other.created_at

  def has_settled(self):
    return True

  def updated_at(self) -> str:
    return self.annotations.get('updated_at') or \
           self.created_at

  def short_desc(self):
    return self.annotations.get('short_desc')

# --
# --
# --
# -------------------------------ACTION-------------------------------
# --
# --
# --

  def reload(self) -> Optional[KR]:
    self.raw = self.find_raw(self.name, self.ns)
    return self if self.raw else None

  @classmethod
  def list(cls, ns=None, **query) -> List[KR]:
    from k8kat.res.relation.relation import Relation
    return Relation[cls](
      model_class=cls,
      ns=ns,
      **query
    )

  @classmethod
  def list_excluding_sys(cls, ns=None, **query) -> List[KR]:
    return cls.list(ns, **query)

  @classmethod
  def wait_until_exists(cls, name: str, ns: str = None):
    res = None
    for attempts in range(0, 20):
      res = cls.find(name, ns)
      if res:
        break
      else:
        time.sleep(1)
    return res

  def delete(self, wait_until_gone=False):
    self._perform_delete_self()
    if wait_until_gone:
      while self.reload():
        time.sleep(0.5)

  def patch(self, modifier=None) -> Optional[KR]:
    if modifier is not None:
      self._enter_patch_loop(modifier)
    else:
      self._perform_patch_self()
    return self.reload()

  def wait_until(self, predicate, max_time_sec=None) -> bool:
    start_time = time.time()
    condition_met = False
    for attempts in range(0, 50):
      if predicate():
        condition_met = True
        break
      else:
        if max_time_sec and time.time() - start_time > max_time_sec:
          return False
        time.sleep(1)
        self.reload()
    return condition_met

  def events(self):
    api = broker.coreV1
    raw_list = api.list_namespaced_event(namespace=self.ns).items
    kat_list = [KatEvent(raw_event) for raw_event in raw_list]
    return [event for event in kat_list if event.is_for(self)]

  def trigger(self):
    self.annotate(trigger=utils.rand_str())

  def touch(self, save=True):
    self.annotate(save=save, updated_at=str(datetime.now()))

  def annotate(self, save=True, **annotations):
    def perf(raw):
      existing = raw.metadata.annotations or {}
      combined = {**existing, **annotations}
      raw.metadata.annotations = combined

    self.patch(perf) if save else perf(self.raw)

  def label(self, save=True, **labels):
    def perf(raw):
      existing = raw.metadata.labels or {}
      combined = {**existing, **labels}
      raw.metadata.labels = combined

    self.patch(perf) if save else perf(self.raw)

# --
# --
# --
# -------------------------------CLASS-------------------------------
# --
# --
# --

  @classmethod
  def find_raw(cls, name, ns=None):
    try:
      fn: Callable = cls.k8s_verb_methods().get('read')
      is_ns: bool = cls.is_namespaced()
      return fn(name=name, namespace=ns) if is_ns else fn(name=name)
    except ApiException:
      return None

  @classmethod
  def find(cls, name, ns=None):
    raw_res = cls.find_raw(name, ns)
    return cls(raw_res) if raw_res else None

  @classmethod
  def delete_if_exists(cls, ns, name, wait_until_gone=False):
    instance = cls.find(name, ns)
    if instance:
      instance.delete(wait_until_gone)

  @classmethod
  def k8s_verb_methods(cls) -> Dict[str, Callable]:
    return dict()

  @classmethod
  def is_namespaced(cls) -> bool:
    return True

  @classmethod
  def inflate(cls, raw) -> KR:
    host = cls.find_res_class(raw.kind)
    return host(raw) if host else None

  @staticmethod
  def find_res_class(kind: str) -> Optional[Type[KR]]:
    subclasses = res_utils.kat_classes()
    predicate = lambda s: kind in s.kind_aliases
    return next(filter(predicate, subclasses), None)

  @classmethod
  def class_for(cls, kind: str) -> Optional[Type[KR]]:
    expl_class = cls.find_res_class(kind)
    return expl_class or auto_namespaced_kat_cls(kind)

# --
# --
# --
# -------------------------------USAGE / REQ / LIM-------------------------------
# --
# --
# --

  def load_per_pod_metrics(self) -> Optional[List[PodMetricsDict]]:
    """Loads the appropriate metrics dict from k8s metrics API."""
    return None

  def cpu_used(self) -> float:
    """Returns resource's total CPU usage in cores or 0"""
    return self._resource_usage('cpu')

  def mem_used(self) -> float:
    """Returns resource's total memory usage in bytes or 0"""
    return self._resource_usage('memory')

  def _resource_usage(self, resource_type: str) -> Optional[float]:
    per_pod_metrics: List[PodMetricsDict] = self.load_per_pod_metrics()
    if per_pod_metrics is not None:
      total = 0
      for pod_metrics in per_pod_metrics:
        per_container_metrics = pod_metrics.get('containers', [])
        for container_metrics in per_container_metrics:
          container_metrics = container_metrics or {}
          metric_deep_key = ('usage', resource_type)
          usage_quant_expr = utils.deep_get(container_metrics, *metric_deep_key)
          usage_quant = units.parse_quant_expr(usage_quant_expr)
          total += usage_quant or 0
      return round(total, 3)
    else:
      return None

  @lru_cache(maxsize=128)
  def pods(self):
    return []

  def _sum_pod_req_or_lim(self, func):
    bytes_per_pod = [func(pod) for pod in self.pods()]
    bytes_per_pod = [value or 0 for value in bytes_per_pod]
    return sum(bytes_per_pod)

  def pods_cpu_limit(self) -> Optional[float]:
    return self._sum_pod_req_or_lim(lambda p: p.cpu_limit())

  def pods_cpu_request(self) -> Optional[float]:
    return self._sum_pod_req_or_lim(lambda p: p.cpu_request())

  def pods_mem_limit(self) -> Optional[float]:
    return self._sum_pod_req_or_lim(lambda p: p.mem_limit())

  def pods_mem_request(self) -> Optional[float]:
    return self._sum_pod_req_or_lim(lambda p: p.mem_request())

# --
# --
# --
# -------------------------------PLUMBING-------------------------------
# --
# --
# --

  def _perform_patch_self(self):
    patch_method = self.k8s_verb_methods().get('patch')
    self.ns_agnostic_call(patch_method, body=self.raw)

  def _enter_patch_loop(self, modification_delegate: Callable):
    failed_attempts = 0
    while True:
      try:
        modification_delegate(self.raw)
        self._perform_patch_self()
        return
      except kubernetes.client.rest.ApiException as e:
        if failed_attempts >= 5:
          raise e
        else:
          failed_attempts += 1
          time.sleep(1)
          self.reload()

  def _perform_delete_self(self):
    impl = self.k8s_verb_methods().get('delete')
    self.ns_agnostic_call(impl)

  def ns_agnostic_call(self, impl: Callable, **kwargs) -> any:
    if self.is_namespaced():
      return impl(name=self.name, namespace=self.ns, **kwargs)
    else:
      return impl(name=self.name, **kwargs)

  def serialize(self, serializer):
    return serializer(self)


def auto_namespaced_kat_cls(res_name_or_kind: str) -> Optional[Type[KatRes]]:
  definition = api_defs_man.find_def(res_name_or_kind)
  if definition:
    class NsdKatShell(KatRes):
      @classproperty
      def kind(self) -> str:
        return definition['kind']

      @classproperty
      def res_name_plural(self) -> str:
        return definition['name']

      @classmethod
      def is_namespaced(cls) -> bool:
        return True

      @classmethod
      def k8s_verb_methods(cls) -> Dict[str, Callable]:
        def _list(namespace, **kwargs):
          return rest_backend.list_namespaced_resources(
            definition['name'],
            definition['apigroup'],
            namespace,
            **kwargs
          )

        def _delete(name, namespace, **kwargs):
          return rest_backend.delete_namespaced_resource(
            definition['name'],
            definition['apigroup'],
            namespace,
            name,
            **kwargs
          )

        return dict(
          list=_list,
          delete=_delete
        )
    return NsdKatShell
  else:
    return None


def from_dict(dict_repr: Dict):
  api = ApiClient()
  mocked_kube_http_resp = FakeKubeResponse(dict_repr)
  kind = dict_repr['kind']
  return api.deserialize(mocked_kube_http_resp, f"V1{kind}")


class FakeKubeResponse:
  def __init__(self, obj):
    import json
    self.data = json.dumps(obj)



import json
import os
import random
import string
import subprocess
from functools import reduce
from pathlib import Path
from typing import Dict

legal_envs = ['production', 'development', 'test']


def shell_exec(cmd) -> str:
  formatted_cmd = cmd.split(' ')
  output = subprocess.run(
    formatted_cmd,
    stdout=subprocess.PIPE
  )
  return output.stdout.decode()


def jk_exec(cmd, **kwargs) -> Dict[str, any]:
  cmd = f"{kmd(cmd, **kwargs)} -o json"
  result = shell_exec(cmd)
  return json.loads(result)


def k_exec(cmd, **kwargs) -> str:
  cmd = kmd(cmd, **kwargs)
  return shell_exec(cmd)


def kmd(cmd: str, **kwargs) -> str:
  kubectl = kwargs.get('k', 'kubectl')
  ns = kwargs.get('ns', None)
  context = kwargs.get('ctx', None)

  cmd = f"{cmd} -n {ns}" if ns else cmd
  cmd = f"{cmd} --context={context}" if context else cmd
  cmd = f"{kubectl} {cmd}"

  return cmd


def root_path() -> str:
  return str(Path(__file__).parent.parent.parent)


def set_run_env(_run_env):
  if _run_env in legal_envs:
    os.environ['KAT_ENV'] = _run_env
  else:
    raise Exception(f"Bad environment '{_run_env}'")


def run_env() -> str:
  return os.environ.get('KAT_ENV', 'production')


def is_prod() -> bool:
  return run_env() == 'production'


def is_dev() -> bool:
  return run_env() == 'development'


def is_test() -> bool:
  return run_env() == 'test'


def is_ci() -> bool:
  return is_test() and os.environ.get('CI')


def is_ci_keep():
  return os.environ.get("CI") == 'keep'


def is_non_trivial(dict_array):
  if not dict_array:
    return False
  return [e for e in dict_array if e]


# noinspection PyBroadException
def try_or(lam, fallback=None):
  try:
    return lam()
  except:
    return fallback


def rand_str(string_len=10):
  letters = string.ascii_lowercase
  return ''.join(random.choice(letters) for i in range(string_len))


def flatten(nested_list):
  return [item for sublist in nested_list for item in sublist]


def deep_get(dictionary, *keys):
  return reduce(lambda d, key: d.get(key) if d else None, keys, dictionary)


def _deep_merge(dict1, dict2):
  for k in set(dict1.keys()).union(dict2.keys()):
    if k in dict1 and k in dict2:
      if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
        yield k, dict(_deep_merge(dict1[k], dict2[k]))
      else:
        # If one of the values is not a dict, you can't continue merging it.
        # Value from second dict overrides one in first and we move on.
        yield k, dict2[k]
        # Alternatively, replace this with exception raiser to alert you of value conflicts
    elif k in dict1:
      yield k, dict1[k]
    else:
      yield k, dict2[k]

def deep_merge(dict1, dict2):
  merged = _deep_merge(dict1, dict2)
  return dict(merged)



from functools import lru_cache
from http.client import HTTPResponse
from typing import List, Optional, Callable, TypeVar, Any

from kubernetes import stream as k8s_streaming
from kubernetes.client import V1Pod, V1Container, \
  V1ContainerState
from kubernetes.client.rest import ApiException

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes, MetricsDict
from k8kat.res.pod import pod_utils
from k8kat.res.pod.pod_utils import when_running_normally
from k8kat.utils.main import utils
from k8kat.utils.main.class_property import classproperty

Fn = TypeVar('Fn', bound=Callable[..., Any])
KP = TypeVar('KP')


class KatPod(KatRes):
  def __init__(self, raw, wait_until_running=False):
    super().__init__(raw)
    if wait_until_running:
      self.wait_until_running()

# --
# --
# --
# -------------------------------PROPERTIES-------------------------------
# --
# --
# --

  @classproperty
  def kind(self):
    return "Pod"

  @property
  def labels(self):
    base = super().labels
    bad_key = 'pod-template-hash'
    return {k: base[k] for k in base.keys() if k != bad_key}

  @property
  def phase(self):
    return self.body().status.phase

  @property
  def ip(self) -> str:
    return utils.try_or(lambda: self.body().status.pod_ip)

  @property
  def has_parent(self) -> bool:
    refs = self.raw.metadata.owner_references
    return refs is not None and len(refs) > 0

# --
# --
# --
# -------------------------------INTEL-------------------------------
# --
# --
# --

  def body(self) -> V1Pod:
    return self.raw

  def container(self, index=0) -> V1Container:
    return self.body().spec.containers[index]

  def image(self, index=0) -> str:
    return self.container and self.container(index).image

# --
# --
# --
# -------------------------------STATE-------------------------------
# --
# --
# --

  def ternary_status(self) -> str:
    if self.is_working():
      return 'positive'
    elif self.is_broken():
      return 'negative'
    else:
      return 'pending'

  def is_running_normally(self):
    if self.is_running():
      main_states = self.main_container_states()
      runners = filter_states(main_states, 'running')
      return len(main_states) == len(runners)

  def is_pending_normally(self):
    return self.is_pending() and \
           not self.is_pending_morbidly()

  def is_running_morbidly(self):
    return self.is_running() and \
           not self.is_running_normally()

  def is_working(self):
    return self.is_running_normally() or \
           self.has_succeeded()

  def is_broken(self) -> bool:
    return self.is_pending_morbidly() or \
           self.is_running_morbidly() or \
           self.has_failed()

  def has_settled(self) -> bool:
    return not self.is_pending_normally()

  def is_pending_morbidly(self) -> bool:
    """Whether this pod is pending because one or more container is failing to start."""
    if self.is_pending():
      init_states = self.init_container_states()
      waiting_init = filter_states(init_states, 'waiting')
      if has_morbid_pending_reasons(waiting_init):
        return True

      main_states = self.main_container_states()
      main_init = filter_states(main_states, 'waiting')
      return has_morbid_pending_reasons(main_init)
    else:
      return False

  def is_running(self) -> bool:
    """Whether this pod is in the Running state"""
    return self.body().status.phase == 'Running'

  def is_pending(self) -> bool:
    """Whether this pod is in the Pending state"""
    return self.body().status.phase == 'Pending'

  def has_failed(self) -> bool:
    """Whether this pod is in the Failed state"""
    return self.body().status.phase == 'Failed'

  def has_succeeded(self):
    """Whether this pod is in the Succeeded state"""
    return self.body().status.phase == 'Succeeded'

  def has_run(self) -> bool:
    """Whether this pod is in the Failed or Succeeded state"""
    return self.has_failed() or self.has_succeeded()

  def raw_logs(self, seconds=60) -> Optional[str]:
    try:
      return broker.coreV1.read_namespaced_pod_log(
        namespace=self.namespace,
        name=self.name,
        since_seconds=seconds
      )
    except ApiException:
      return None

  def clean_logs(self, seconds=60) -> str:
    _raw_logs = self.raw_logs(seconds)
    return (_raw_logs or '').strip()

  def log_lines(self, seconds=60) -> List[str]:
    raw_log_str = self.raw_logs(seconds)
    if raw_log_str:
      return raw_log_str.strip("\n").split("\n")
    else:
      return []

# --
# --
# --
# ---------------------------REQ / LIM / METRICS---------------------------
# --
# --
# --

  def cpu_request(self) -> Optional[float]:
    """Returns pod's total memory limits in bytes."""
    return self.read_resources_req_or_lim('requests', 'cpu')

  def cpu_limit(self) -> Optional[float]:
    """Returns pod's total memory limits in bytes."""
    return self.read_resources_req_or_lim('limits', 'cpu')

  def mem_limit(self) -> Optional[float]:
    """Returns pod's total memory limits in bytes."""
    return self.read_resources_req_or_lim('limits', 'memory')

  def mem_request(self) -> Optional[float]:
    """Returns pod's total memory requests in bytes."""
    return self.read_resources_req_or_lim('requests', 'memory')

  def eph_storage_limit(self) -> Optional[float]:
    """Returns pod's total ephemeral storage limits in bytes."""
    return self.read_resources_req_or_lim('limits', 'ephemeral-storage')

  def eph_storage_request(self) -> Optional[float]:
    """Returns pod's total ephemeral storage requests in bytes."""
    return self.read_resources_req_or_lim('requests', 'ephemeral-storage')

  def read_resources_req_or_lim(self, metric: str, resource: str) -> Optional[float]:
    """Fetches pod's total resource capacity (either limits or requests)
    for either CPU (cores) or memory (bytes)."""
    container_lvl_value = lambda c: pod_utils.container_req_or_lim(c, metric, resource)
    containers = self.body().spec.containers or []
    per_container_results = [container_lvl_value(c) for c in containers]
    are_all_undefined = len([v for v in per_container_results if v is not None]) == 0
    return sum(per_container_results) if not are_all_undefined else None

  @when_running_normally
  @lru_cache(maxsize=128)
  def load_per_pod_metrics(self) -> Optional[List[MetricsDict]]:
    """Loads the appropriate metrics dict from k8s metrics API."""
    try:
      self_metrics = broker.custom.get_namespaced_custom_object(
        group='metrics.k8s.io',
        version='v1beta1',
        namespace=self.namespace,
        plural='pods',
        name=self.name
      )
      return [self_metrics] if self_metrics else None
    except ApiException:
      return []

# --
# --
# --
# -------------------------------ACTION-------------------------------
# --
# --
# --

  @classmethod
  def consume_runner(cls, name: str, ns: str, wait_until_gone=False) -> Optional[str]:
    pod = cls.wait_until_exists(name, ns)
    if pod:
      pod.wait_until(pod.has_run)
      logs = None
      if pod.has_succeeded():
        logs = pod.raw_logs()
      pod.delete(wait_until_gone)
      return logs
    else:
      return None

  @when_running_normally
  def shell_exec(self, command) -> Optional[str]:
    fmt_command = pod_utils.coerce_cmd_format(command)
    result = k8s_streaming.stream(
      broker.coreV1.connect_get_namespaced_pod_exec,
      self.name,
      self.namespace,
      command=fmt_command,
      stderr=True,
      stdin=False,
      stdout=True,
      tty=False
    )
    return result.strip() if result else None

  def replace_image(self, new_image_name, index=0):
    self.body().spec.containers[index].image = new_image_name
    self.patch()

  def wait_until_running(self):
    return self.wait_until(self.is_running)

  def invoke_curl(self, **kwargs) -> HTTPResponse:
    fmt_command = pod_utils.build_curl_cmd(**kwargs, with_command=True)
    result = self.shell_exec(fmt_command)
    if result is not None:
      result = pod_utils.parse_response(result)
    return result

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespaced_pod,
      patch=broker.coreV1.patch_namespaced_pod,
      delete=broker.coreV1.delete_namespaced_pod,
      list=broker.coreV1.list_namespaced_pod
    )

# --
# --
# --
# -------------------------------UTILS-------------------------------
# --
# --
# --

  def main_container_states(self) -> List[V1ContainerState]:
    statuses = self.body().status.container_statuses or []
    return [status.state for status in statuses]

  def init_container_states(self) -> List[V1ContainerState]:
    statuses = self.body().status.init_container_statuses or []
    return [status.state for status in statuses]


def has_morbid_pending_reasons(states: List[V1ContainerState]):
  stated_reasons = set([state.waiting.reason for state in states])
  good_reasons = {'ContainerCreating', 'PullingImage', 'PodInitializing'}
  bad_reasons = stated_reasons - good_reasons
  return len(bad_reasons) > 0


def filter_states(states: List[V1ContainerState], _type: str) -> List[V1ContainerState]:
  return [state for state in states if getattr(state, _type)]



from typing import Optional, Dict

from kubernetes.client import V1ResourceQuota

from k8kat.auth.kube_broker import broker
from k8kat.res.base.kat_res import KatRes
from k8kat.utils.main import units
from k8kat.utils.main.class_property import classproperty


class KatQuota(KatRes):

  @classproperty
  def kind(self):
    return "ResourceQuota"

  def body(self) -> V1ResourceQuota:
    return self.raw

  def dump(self):
    output = {}
    raw_spec_hard = self.extract_values('spec', 'hard')
    raw_status_used = self.extract_values('status', 'used')
    for key in raw_spec_hard.keys():
      output[key] = dict(
        spec=units.parse_quant_expr(raw_spec_hard[key]),
        used=units.parse_quant_expr(raw_status_used.get(key))
      )
    return output

  def mem_request_sum_allowed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('spec', 'hard', 'memory')
    return units.parse_quant_expr(expr, target_unit)

  def mem_request_sum_deployed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('status', 'used', 'memory')
    return units.parse_quant_expr(expr, target_unit)

  def mem_limit_sum_allowed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('spec', 'hard', 'limits.memory')
    return units.parse_quant_expr(expr, target_unit)

  def mem_limit_sum_deployed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('status', 'used', 'limits.memory')
    return units.parse_quant_expr(expr, target_unit)

  def cpu_request_sum_allowed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('spec', 'hard', 'cpu')
    return units.parse_quant_expr(expr, target_unit)

  def cpu_request_sum_deployed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('status', 'used', 'cpu')
    return units.parse_quant_expr(expr, target_unit)

  def cpu_limit_sum_allowed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('spec', 'hard', 'limits.cpu')
    return units.parse_quant_expr(expr, target_unit)

  def cpu_limit_sum_deployed(self, target_unit='') -> Optional[float]:
    expr = self.extract_value('status', 'used', 'limits.cpu')
    return units.parse_quant_expr(expr, target_unit)

  def extract_value(self, source_name, source_key, value_key) -> str:
    source = getattr(self.body(), source_name, None)
    values = getattr(source, source_key, {}) if source else {}
    result = (values or {}).get(value_key, '')
    if not result and 'requests.' not in value_key:
      result = (values or {}).get(f"requests.{value_key}", '')
    return result

  def extract_values(self, source_name, source_key) -> Dict[str, str]:
    source = getattr(self.body(), source_name, None)
    return getattr(source, source_key, {}) if source else {}

  @classmethod
  def list_excluding_sys(cls, ns=None, **query):
    updated_query = dict(
      **query,
      not_fields={
        **(query.get('not_fields', {})),
        'metadata.name': 'gke-resource-quotas',
      }
    )
    return cls.list(ns, **updated_query)

  @classmethod
  def k8s_verb_methods(cls):
    return dict(
      read=broker.coreV1.read_namespaced_resource_quota,
      patch=broker.coreV1.patch_namespaced_resource_quota,
      delete=broker.coreV1.delete_namespaced_resource_quota,
      list=broker.coreV1.list_namespaced_resource_quota
    )



from typing import Optional, Tuple, Dict, Union


def find_multiplier_mapping(expr: str) -> Optional[Tuple[str, Union[int,float]]]:
  """Finds the correct multiplier for unit conversion."""
  if expr is None:
    return None
  for candidate, multiplier in unit_map.items():
    if expr.endswith(candidate):
      return candidate, multiplier
  return None

def parse_quant_expr(src_expr: str, target: str = '') -> Optional[float]:
  """Parses the source expression into standard units
  or target units (if specified)."""
  if src_expr is not None:
    src_mapping = find_multiplier_mapping(src_expr)
    target_mapping = find_multiplier_mapping(target)
    if not src_mapping or not target_mapping:
      return None

    try:
      src_quantity, src_multiplier = src_mapping
      _, target_multiplier = target_mapping
      unitless = float(src_expr.strip(src_quantity))
      return unitless * src_multiplier / target_multiplier
    except ValueError:
      return None
  else:
    return None

def humanize_cpu_quant(cores: float, with_unit: bool = False) -> str:
  """Returns CPU quantity in cores rounded to single decimal."""
  base = "{:.1f}".format(cores)
  return f"{base} {'Cores' if with_unit else ''}".strip(' ')


def humanize_mem_quant(byte_value: float) -> Optional[str]:
  """ Returns memory quantity converted from bytes to higher level units.
  Automatically picks the right units for friendliest display."""
  unit_map_items = list(unit_map.items())
  sorted_items = sorted(unit_map_items, key=lambda item: item[1])
  unit, bytes_in_unit = None, None
  for (try_unit, multiplier) in sorted_items:
    bytes_in_unit = byte_value / multiplier
    if try_unit and bytes_in_unit < 1000:
      unit = try_unit
      break
  return f"{int(bytes_in_unit)}{unit}b" if bytes_in_unit else None


unit_map: Dict[str, Union[int, float]] = {
  'Ei'   : 2  ** 60,
  'Pi'   : 2  ** 50,
  'Ti'   : 2  ** 40,
  'Gi'   : 2  ** 30,
  'Mi'   : 2  ** 20,
  'Ki'   : 2  ** 10,
  'E'    : 10 ** 18,
  'P'    : 10 ** 15,
  'T'    : 10 ** 12,
  'G'    : 10 ** 9,
  'M'    : 10 ** 6,
  'K'    : 10 ** 3,
  'k'    : 10 ** 3,  # k8s uses both K and k for "kilo"
  'm'    : 10 ** -3,
  'micro': 10 ** -6,
  'n'    : 10 ** -9,
  'p'    : 10 ** -12,
  ''     : 1
}